import sys
from pathlib import Path
from typing import TYPE_CHECKING, Generic, TypeVar

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

if TYPE_CHECKING:
    from sigmatcher.definitions import MacroStatement, Signature

SignatureMatch = TypeVar("SignatureMatch", str, Path)


class SigmatcherError(Exception):
    def __init__(self, analyzer_name: str, *args: object) -> None:
        self.analyzer_name: str = analyzer_name
        super().__init__(analyzer_name, *args)

    def debug_message(self) -> str:
        return ""

    def short_message(self) -> str:
        return str(self)


class SignaturesCountError(SigmatcherError):
    pass


class NoSignaturesError(SignaturesCountError):
    @override
    def short_message(self) -> str:
        return "Found no signatures! Make sure your version ranges are correct."


class TooManySignaturesError(SignaturesCountError):
    def __init__(self, analyzer_name: str, signatures: "tuple[Signature, ...]", *args: object) -> None:
        self.signatures: tuple[Signature, ...] = signatures
        super().__init__(analyzer_name, signatures, *args)

    @override
    def short_message(self) -> str:
        return f"Found {len(self.signatures)} signatures. Field definitions should only have one."


class MatchError(SigmatcherError):
    def __init__(self, analyzer_name: str, signatures: "tuple[Signature,...] | None", *args: object) -> None:
        self.signatures: tuple[Signature, ...] | None = signatures
        super().__init__(analyzer_name, signatures, *args)

    @override
    def debug_message(self) -> str:
        if self.signatures is None:
            return ""

        signature_mesage = "\n ".join(signature.model_dump_json(exclude_defaults=True) for signature in self.signatures)
        return f"- Signatures:\n {signature_mesage}"


class NoMatchesError(MatchError):
    @override
    def short_message(self) -> str:
        return "Found no matches!"


class TooManyMatchesError(MatchError, Generic[SignatureMatch]):
    def __init__(
        self,
        analyzer_name: str,
        matches: "set[SignatureMatch]",
        signatures: "tuple[Signature,...] | None",
        *args: object,
    ) -> None:
        self.matches: set[SignatureMatch] = matches
        super().__init__(analyzer_name, signatures, matches, *args)

    @override
    def debug_message(self) -> str:
        matches_message = "\n ".join(str(match) for match in self.matches)
        return f"{super().debug_message()}\n- Matches:\n {matches_message}"

    @override
    def short_message(self) -> str:
        return "Found too many matches"


class DependencyMatchError(SigmatcherError):
    def __init__(self, analyzer_name: str, missing_dependencies: list[str], *args: object) -> None:
        self.missing_dependencies: list[str] = missing_dependencies
        self.should_show_debug: bool = True
        super().__init__(analyzer_name, missing_dependencies, *args)

    @override
    def short_message(self) -> str:
        return "Skipped because of failed dependencies"

    @override
    def debug_message(self) -> str:
        if not self.should_show_debug:
            return ""
        dependencies_message = "\n ".join(self.missing_dependencies)
        return f"- Dependecies: \n{dependencies_message}"


class InvalidMacroModifierError(SigmatcherError):
    """
    Exception raised when an invalid macro modifier is encountered.
    """

    def __init__(self, analyzer_name: str, macro: "MacroStatement", result_class_name: str, *args: object) -> None:
        self.macro: MacroStatement = macro
        self.result_class_name: str = result_class_name
        super().__init__(analyzer_name, macro, result_class_name, *args)

    @override
    def debug_message(self) -> str:
        return (
            f"Macro Subject: {self.macro.subject}\nMacro Modifier: {self.macro.modifier}\n"
            + f"Subject Class: {self.result_class_name}"
        )

    @override
    def short_message(self) -> str:
        return "Invalid macro modifier"



import fnmatch
import re
import sys
from abc import ABC, abstractmethod
from collections.abc import Iterable, Sequence
from dataclasses import dataclass
from pathlib import Path
from typing import Annotated, ClassVar, Literal, TypeAlias, TypeVar

import pydantic
from packaging.specifiers import SpecifierSet

from sigmatcher.grep import rip_regex

if sys.version_info < (3, 11):
    from typing_extensions import Self
else:
    from typing import Self

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override


def is_in_version_range(app_version: str | None, version_range: str | list[str] | None) -> bool:
    if app_version is None or version_range is None:
        return True
    ranges = version_range if isinstance(version_range, list) else [version_range]
    return any(SpecifierSet(spec).contains(app_version) for spec in ranges)


@dataclass(frozen=True)
class MacroStatement:
    subject: str
    modifier: str


class CountRange(pydantic.BaseModel):
    min_count: int
    max_count: int

    def __contains__(self, value: int) -> bool:
        return self.min_count <= value <= self.max_count


class BaseSignature(ABC, pydantic.BaseModel, frozen=True, use_attribute_docstrings=True, extra="forbid"):  # pyright: ignore[reportUnsafeMultipleInheritance]
    version_range: str | list[str] | None = None
    """The version range in which the signature is valid."""
    count: Annotated[CountRange, pydantic.Field(json_schema_extra={"default": 1})] = CountRange(
        min_count=1, max_count=1
    )
    """The number of times the signature should match in order to be considered a match.
    Can be either an integer or a string of the form "min-max"."""

    @abstractmethod
    def check_files(self, search_paths: set[Path], search_root: Path) -> list[Path]:
        raise NotImplementedError()

    @abstractmethod
    def check_strings(self, strings: Iterable[str]) -> list[str]:
        raise NotImplementedError()

    @abstractmethod
    def capture(self, value: str) -> set[str]:
        raise NotImplementedError()

    @abstractmethod
    def get_dependencies(self) -> list[str]:
        raise NotImplementedError()

    @abstractmethod
    def get_macro_definitions(self) -> set[MacroStatement]:
        raise NotImplementedError()

    @abstractmethod
    def resolve_macro(self, macro_statement: MacroStatement, resolved_macro: str) -> Self:
        raise NotImplementedError()

    def is_in_version_range(self, app_version: str | None) -> bool:
        return is_in_version_range(app_version, self.version_range)

    @pydantic.field_validator("count", mode="before", json_schema_input_type=int | str)
    @classmethod
    def _parse_count(cls, v: str | int) -> CountRange:
        if isinstance(v, str):
            min_count, max_count = map(int, v.split("-"))
        else:
            min_count, max_count = v, v
        return CountRange(min_count=min_count, max_count=max_count)

    @pydantic.field_serializer("count", mode="plain")
    @staticmethod
    def _serialize_count(count: CountRange) -> int | str:
        if count.min_count == count.max_count:
            return count.min_count
        return f"{count.min_count}-{count.max_count}"


class BaseRegexSignature(BaseSignature, frozen=True):
    signature: "re.Pattern[str]" = pydantic.Field(
        json_schema_extra={"x-intellij-language-injection": {"language": "RegExp"}}
    )
    """
    A regular expression used to check the signature.

    When used for capturing strings, if there is more than one match,
    a named group called `match` should be used to specify what should be captured.

    Can include macros in the form of `${<result>.<property>}`.
    Results are actually objects from `sigmatcher.results` package.
    i.e. `Class`, `Field`, `Method`, `Export` objects.
    Property could be any python property that the objects holds.

    For example a macro could look like `${MyClass.fields.java}`, which would return something like
    `Lf9/s;->a:LX/Y/Z/A02;`
    """

    MACRO_REGEX: "ClassVar[re.Pattern[str]]" = re.compile(r"\${(.*?)}")

    @override
    def check_files(self, search_paths: set[Path], search_root: Path) -> list[Path]:
        # Limit the search to avoid too many arguments to ripgrep
        match_limit = 100
        if len(search_paths) > match_limit:
            search_paths_for_rg = {search_root}
        else:
            search_paths_for_rg = search_paths
        file_to_match_count = rip_regex(self.signature, search_paths_for_rg)
        if self.count.min_count > 0:
            # Optimization: do not iterate over files with 0 matches
            return [path.resolve() for path, match_count in file_to_match_count.items() if match_count in self.count]
        # rip_regex only includes files with count > 0 in the result.
        return [path.resolve() for path in search_paths if file_to_match_count.get(path, 0) in self.count]

    @override
    def check_strings(self, strings: Iterable[str]) -> list[str]:
        results: list[str] = []
        for string in strings:
            match_count = len(self.signature.findall(string))
            if match_count in self.count:
                results.append(string)
        return results

    @override
    def capture(self, value: str) -> set[str]:
        match = self.signature.search(value)
        if match is None:
            return set()
        try:
            return {match.group("match")}
        except IndexError:
            pass

        return set(match.groups())

    @override
    def get_dependencies(self) -> list[str]:
        return [macro.subject for macro in self.get_macro_definitions()]

    @override
    def get_macro_definitions(self) -> set[MacroStatement]:
        macros: set[MacroStatement] = set()

        for raw_macro in self.MACRO_REGEX.findall(self.signature.pattern):  # pyright: ignore[reportAny]
            assert isinstance(raw_macro, str)
            macro_subject, _, macro_modifier = raw_macro.rpartition(".")
            macros.add(MacroStatement(macro_subject, macro_modifier))

        return macros

    @override
    def resolve_macro(self, macro_statement: MacroStatement, resolved_macro: str) -> Self:
        macro_string = f"{macro_statement.subject}.{macro_statement.modifier}"
        new_pattern = self.signature.pattern.replace(f"${{{macro_string}}}", re.escape(resolved_macro))

        return self.model_copy(update={"signature": re.compile(new_pattern)})


class RegexSignature(BaseRegexSignature, frozen=True):
    type: Literal["regex"] = "regex"
    """The type of the signature."""


class GlobSignature(BaseRegexSignature, frozen=True):
    type: Literal["glob"] = "glob"
    """The type of the signature."""

    @pydantic.field_validator("signature", mode="before")
    @classmethod
    def parse_glob(cls, v: str) -> str:
        # TODO: removing the atomic group, i.e. the "(?>" makes glob signature in the form of "*WORD*" slower then
        #  their regex counterparts
        return fnmatch.translate(v).replace("\\Z", "$").replace("(?>", "(?:")


class TreeSitterSignature(BaseSignature, frozen=True):
    signature: str
    """A TreeSitter s-query used to check the signature."""
    type: Literal["treesitter"] = "treesitter"
    """The type of the signature."""

    @override
    def check_files(self, search_paths: set[Path], search_root: Path) -> list[Path]:
        raise NotImplementedError("TreeSitter signatures are not supported yet.")

    @override
    def check_strings(self, strings: Iterable[str]) -> list[str]:
        raise NotImplementedError("TreeSitter signatures are not supported yet.")

    @override
    def capture(self, value: str) -> set[str]:
        raise NotImplementedError("TreeSitter signatures are not supported yet.")

    @override
    def get_dependencies(self) -> list[str]:
        raise NotImplementedError("TreeSitter signatures are not supported yet.")

    @override
    def get_macro_definitions(self) -> set[MacroStatement]:
        raise NotImplementedError("TreeSitter signatures are not supported yet.")

    @override
    def resolve_macro(self, macro_statement: MacroStatement, resolved_macro: str) -> Self:
        raise NotImplementedError("TreeSitter signatures are not supported yet.")


Signature: TypeAlias = Annotated[
    RegexSignature | GlobSignature | TreeSitterSignature, pydantic.Field(discriminator="type")
]

SignatureMatch = TypeVar("SignatureMatch", str, Path)


class Definition(pydantic.BaseModel, frozen=True, use_attribute_docstrings=True, extra="forbid"):
    name: str
    """The name of the definition, i.e. the class, method, field, or export name."""
    signatures: tuple[Signature, ...]
    """A list of signatures that define the definition."""
    version_range: str | list[str] | None = None
    """The version range in which the definition is valid."""

    def get_signatures_for_version(self, app_version: str | None) -> tuple[Signature, ...]:
        if app_version is None:
            return self.signatures
        return tuple(signature for signature in self.signatures if signature.is_in_version_range(app_version))

    def get_dependencies(self, app_version: str | None) -> set[str]:
        dependencies: set[str] = set()
        for signature in self.get_signatures_for_version(app_version):
            dependencies.update(signature.get_dependencies())
        return dependencies

    def is_in_version_range(self, app_version: str | None) -> bool:
        return is_in_version_range(app_version, self.version_range)


class ExportDefinition(Definition, frozen=True):
    pass


class FieldDefinition(Definition, frozen=True):
    pass


class MethodDefinition(Definition, frozen=True):
    pass


class ClassDefinition(Definition, frozen=True):
    package: str | None = None
    """The package of the class."""
    fields: tuple[FieldDefinition, ...] = ()
    """A list of field definitions."""
    methods: tuple[MethodDefinition, ...] = ()
    """A list of method definitions."""
    exports: tuple[ExportDefinition, ...] = ()
    """A list of export definitions."""


DEFINITIONS_TYPE_ADAPTER = pydantic.TypeAdapter(list[ClassDefinition])


TDefinition = TypeVar("TDefinition", bound=Definition)


def merge_definition(def1: TDefinition, def2: TDefinition) -> TDefinition:
    signatures = def2.signatures if def2.signatures else def1.signatures

    if isinstance(def1, ClassDefinition):
        assert isinstance(def2, ClassDefinition)
        return (
            def1.model_copy(  # pyrefly: ignore[bad-return] related to https://github.com/facebook/pyrefly/issues/1274
                update={
                    "signatures": signatures,
                    "methods": merge_definitions_groups([def1.methods, def2.methods]),
                    "fields": merge_definitions_groups([def1.fields, def2.fields]),
                    "exports": merge_definitions_groups([def1.exports, def2.exports]),
                }
            )
        )
    return def1.model_copy(update={"signatures": signatures})


def merge_definitions_groups(definition_groups: Sequence[Sequence[TDefinition]]) -> tuple[TDefinition, ...]:
    """
    Merge multiple definition groups into a single group,
    e.g. merge all the methods and fields of 2 class definitions with the same name.
    When a definition is present in multiple groups, the signatures are taken from the last group.
    """
    final_definitions: dict[str, TDefinition] = {}
    for definition_group in definition_groups:
        for definition in definition_group:
            if definition.name in final_definitions:
                final_definitions[definition.name] = merge_definition(final_definitions[definition.name], definition)
            else:
                final_definitions[definition.name] = definition
    return tuple(final_definitions.values())



import hashlib
import io
import json
import shutil
import subprocess
import sys
from pathlib import Path
from typing import Annotated

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

import platformdirs
import pydantic
import pydantic.json_schema
import pydantic_core
import rich.markup
import typer
import yaml
from packaging import version
from rich.console import Console, Group, RenderableType
from rich.padding import Padding
from rich.tree import Tree

import sigmatcher.analysis
from sigmatcher import __version__
from sigmatcher.definitions import DEFINITIONS_TYPE_ADAPTER, ClassDefinition, merge_definitions_groups
from sigmatcher.exceptions import DependencyMatchError, SigmatcherError
from sigmatcher.formats import MappingFormat, convert_to_format, parse_from_format
from sigmatcher.results import MatchedClass, Result

app = typer.Typer()

stdout_console = Console()
stderr_console = Console(stderr=True)

cache_app = typer.Typer(help="Manage Sigmatcher's cache.")
app.add_typer(cache_app, name="cache")

CACHE_DIR_PATH = platformdirs.user_cache_path("sigmatcher", "oriori1703", ensure_exists=True)


@cache_app.command()
def info(
    apk: Annotated[
        Path | None,
        typer.Argument(
            help="Optionally, a path to an APK file. If omitted, prints info about the global cache.",
            dir_okay=False,
            exists=True,
        ),
    ] = None,
) -> None:
    """
    Get the path to Sigmatcher's cache directory.
    """
    if apk is None:
        print(str(CACHE_DIR_PATH))
    else:
        print(get_cache_directory(apk))


@cache_app.command()
def clean(
    apk: Annotated[
        Path | None,
        typer.Argument(
            help="Optionally, a path to a specific APK file to clean the cache for. If omitted, cleans everything.",
            dir_okay=False,
            exists=True,
        ),
    ] = None,
) -> None:
    """
    Clean the cache directory.
    """
    if apk is not None:
        shutil.rmtree(get_cache_directory(apk))
    else:
        for path in CACHE_DIR_PATH.iterdir():
            shutil.rmtree(path)
    stdout_console.print("[green]Successfully cleaned the cache directory.[/green]")


def get_cache_directory(apk: Path) -> Path:
    apk_hash_hex = hashlib.sha256(apk.read_bytes()).hexdigest()
    return CACHE_DIR_PATH / apk_hash_hex


def version_callback(value: bool) -> None:
    if value:
        stdout_console.print(f"Sigmatcher version: [green]{rich.markup.escape(__version__)}[/green]")
        raise typer.Exit()


@app.callback()
def callback(
    _version: Annotated[
        bool | None, typer.Option("--version", help="Show the version and exit.", callback=version_callback)
    ] = None,
) -> None:
    pass


@app.command()
def schema(
    output: Annotated[
        Path | None, typer.Option(help="Output path for the json schema. If none is given print to stdout")
    ] = None,
) -> None:
    """
    Get the json schema for writing definitions.
    """

    class SigmatcherGenerateJsonSchema(pydantic.json_schema.GenerateJsonSchema):
        @override
        def generate(
            self, schema: pydantic_core.CoreSchema, mode: pydantic.json_schema.JsonSchemaMode = "validation"
        ) -> pydantic.json_schema.JsonSchemaValue:
            json_schema = super().generate(schema, mode=mode)
            json_schema["title"] = "Sigmatcher's Definitions"
            json_schema["$schema"] = self.schema_dialect
            return json_schema

    definitions_schema = DEFINITIONS_TYPE_ADAPTER.json_schema(schema_generator=SigmatcherGenerateJsonSchema)
    definitions_schema_json = json.dumps(definitions_schema, indent=2)
    if output is not None:
        _ = output.write_text(definitions_schema_json)
    else:
        stdout_console.print(rich.markup.escape(definitions_schema_json))


def apktool_callback(value: str) -> str:
    resolved_path = shutil.which(value)
    if resolved_path is None:
        raise typer.BadParameter("Cannot find the apktool executable")
    return resolved_path


@app.command()
def convert(
    input_file: Annotated[
        Path | None, typer.Option(help="Path for the input mapping output", exists=True, file_okay=True, dir_okay=False)
    ] = None,
    input_format: Annotated[
        MappingFormat, typer.Option(help="The mapping format of the input file")
    ] = MappingFormat.ENIGMA,
    output_file: Annotated[Path | None, typer.Option(help="Path for the output mapping file")] = None,
    output_format: Annotated[MappingFormat, typer.Option(help="The output mapping format")] = MappingFormat.RAW,
) -> None:
    """
    Convert a mapping output from one format to another.
    """
    if input_file is not None:
        raw_input = input_file.read_text()
    elif isinstance(sys.stdin, io.TextIOBase):
        raw_input = sys.stdin.read()
    else:
        raise ValueError("Cannot read from stdin, please provide an input file")
    intermidiate_mappings = parse_from_format(raw_input, input_format)

    mapping_output = convert_to_format(intermidiate_mappings, output_format)
    if output_file is None:
        stdout_console.print(rich.markup.escape(mapping_output))
    else:
        _ = output_file.write_text(mapping_output)


def _read_definitions(signatures: list[Path]) -> tuple[ClassDefinition, ...]:
    definition_groups: list[tuple[ClassDefinition, ...]] = []
    for signature_file in signatures:
        with signature_file.open("r") as f:
            raw_yaml = yaml.safe_load(f)  # pyright: ignore[reportAny]
            definitions = DEFINITIONS_TYPE_ADAPTER.validate_python(raw_yaml)
        definition_groups.append(tuple(definitions))
    return merge_definitions_groups(definition_groups)


def _get_apktool_version(apktool: str) -> str:
    proc = subprocess.run([apktool, "--version"], check=True, capture_output=True)
    return proc.stdout.decode()


def _unpack_apk(apktool: str, apk: Path) -> Path:
    unpacked_path = get_cache_directory(apk)
    if not unpacked_path.exists():
        if version.parse(_get_apktool_version(apktool)) >= version.parse("2.12.0"):
            only_manifest_flags = ["--only-manifest"]
        else:
            only_manifest_flags = ["--no-res", "--force-manifest"]
        _ = subprocess.run(
            [
                apktool,
                "decode",
                apk,
                *only_manifest_flags,
                "--no-assets",
                "-f",
                "--output",
                unpacked_path.with_suffix(".tmp"),
            ],
            check=True,
        )
        _ = shutil.move(unpacked_path.with_suffix(".tmp"), unpacked_path)
    return unpacked_path


def _get_apk_version(unpacked_path: Path) -> str | None:
    apktool_yaml_file = unpacked_path / "apktool.yml"
    with apktool_yaml_file.open() as f:
        apk_version = yaml.safe_load(f)["versionInfo"]["versionName"]  # pyright: ignore[reportAny]

    if isinstance(apk_version, float | int):
        apk_version = str(apk_version)
    assert isinstance(apk_version, str) or apk_version is None
    return apk_version


def _output_successful_results(
    results: dict[str, MatchedClass], output_file: Path | None, output_format: MappingFormat
) -> None:
    mapping_output = convert_to_format(results, output_format)
    if output_file is None:
        stdout_console.print(rich.markup.escape(mapping_output), overflow="ignore", no_wrap=True, crop=False)
    else:
        _ = output_file.write_text(mapping_output)


def _render_error(error: SigmatcherError, debug: bool) -> RenderableType:
    error_message = (
        f"[red]{rich.markup.escape(error.analyzer_name)}[/red] - {rich.markup.escape(error.short_message())}"
    )
    if debug and (debug_message := error.debug_message()):
        formated_debug_message = Padding(f"[yellow]{rich.markup.escape(debug_message)}[/yellow]", (0, 4))
        return Group(error_message, "[yellow]Debug Info:[/yellow]", formated_debug_message)

    return Group(error_message)


def _output_failed_results_flat(failed_results: dict[str, SigmatcherError], debug: bool) -> None:
    stderr_console.print("Errors:")
    for result in failed_results.values():
        stderr_console.print(_render_error(result, debug))


def _output_failed_results_tree(failed_results: dict[str, SigmatcherError], debug: bool) -> None:
    dependent_errors: dict[str, list[SigmatcherError]] = {}
    top_level_errors: list[SigmatcherError] = []
    for result in failed_results.values():
        if isinstance(result, DependencyMatchError):
            result.should_show_debug = False
            for dependecy in result.missing_dependencies:
                dependent_errors.setdefault(dependecy, []).append(result)
        else:
            top_level_errors.append(result)

    def create_error_tree(error: SigmatcherError, tree: Tree) -> None:
        error_message = _render_error(error, debug)
        branch = tree.add(error_message)
        for dependent_error in dependent_errors.get(error.analyzer_name, []):
            create_error_tree(dependent_error, branch)

    tree = Tree("Errors:")
    for result in top_level_errors:
        create_error_tree(result, tree)

    stderr_console.print(tree)


def _output_results(
    results: dict[str, Result | SigmatcherError],
    output_file: Path | None,
    output_format: MappingFormat,
    output_errors_as_tree: bool,
    debug: bool,
) -> None:
    successful_results: dict[str, MatchedClass] = {}
    failed_results: dict[str, SigmatcherError] = {}

    for analyzer_name, result in results.items():
        if isinstance(result, SigmatcherError):
            failed_results[analyzer_name] = result
        elif isinstance(result, MatchedClass):
            successful_results[analyzer_name] = result

    _output_successful_results(successful_results, output_file, output_format)
    if not failed_results:
        return
    if output_errors_as_tree:
        _output_failed_results_tree(failed_results, debug)
    else:
        _output_failed_results_flat(failed_results, debug)


@app.command()
def analyze(  # noqa: PLR0913
    apk: Annotated[
        Path, typer.Argument(help="Path to the apk that will be analyzed", exists=True, file_okay=True, dir_okay=False)
    ],
    signatures: Annotated[
        list[Path],
        typer.Option(
            help="Path to a signature file. If multiple files are given, they are merged together",
            exists=True,
            file_okay=True,
            dir_okay=False,
        ),
    ],
    output_file: Annotated[Path | None, typer.Option(help="Output path for the final mapping output")] = None,
    output_format: Annotated[MappingFormat, typer.Option(help="The output mapping format")] = MappingFormat.RAW,
    tree_errors: Annotated[bool, typer.Option(help="Show dependency errors as a tree")] = False,
    debug: Annotated[
        bool, typer.Option(help="Provide more verbose error messages to help you debug match failures")
    ] = False,
    apktool: Annotated[
        str, typer.Option(help="The command to use when running apktool", callback=apktool_callback)
    ] = "apktool",
) -> None:
    """
    Analyze an APK file using the provided signatures.
    """
    merged_definitions = _read_definitions(signatures)
    unpacked_path = _unpack_apk(apktool, apk)

    apk_version = _get_apk_version(unpacked_path)
    if apk_version is None:
        stderr_console.print("[yellow][Warning][/yellow] No version was found in the APK. Using 0.0.0.0")
        apk_version = "0.0.0.0"

    results = sigmatcher.analysis.analyze(merged_definitions, unpacked_path, apk_version)
    _output_results(results, output_file, output_format, tree_errors, debug)


def main() -> None:
    app()



import enum
import json
import sys
from abc import ABC, abstractmethod
from io import StringIO
from typing import ClassVar, Literal

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

import pydantic
import pydantic.alias_generators

from sigmatcher.results import Class, Field, MatchedClass, MatchedField, MatchedMethod, Method


class Formatter(ABC):
    @abstractmethod
    def convert(self, matched_classes: dict[str, MatchedClass]) -> str:
        raise NotImplementedError()


class Parser(ABC):
    @abstractmethod
    def parse(self, raw_input: str) -> dict[str, MatchedClass]:
        raise NotImplementedError()


class RawFormatter(Formatter):
    @override
    def convert(self, matched_classes: dict[str, MatchedClass]) -> str:
        return pydantic.RootModel[dict[str, MatchedClass]](matched_classes).model_dump_json(indent=4)


class RawParser(Parser):
    @override
    def parse(self, raw_input: str) -> dict[str, MatchedClass]:
        return pydantic.RootModel[dict[str, MatchedClass]].model_validate_json(raw_input).root


class LegacyFormatter(Formatter):
    @override
    def convert(self, matched_classes: dict[str, MatchedClass]) -> str:
        return json.dumps(
            {
                matched_class.original.name: {
                    "className": f"{matched_class.new.package}.{matched_class.new.name}",
                    "methods": {method.original.name: method.new.name for method in matched_class.matched_methods},
                    "fields": {field.original.name: field.new.name for field in matched_class.matched_fields},
                }
                for matched_class in matched_classes.values()
            },
            indent=4,
            sort_keys=True,
        )


class EnigmaFormatter(Formatter):
    def convert_field(self, field: MatchedField) -> str:
        return f"\tFIELD {field.new.name} {field.original.name} {field.new.type}\n"

    def convert_method(self, method: MatchedMethod) -> str:
        return (
            f"\tMETHOD {method.new.name} {method.original.name} ({method.new.argument_types}){method.new.return_type}\n"
        )

    def convert_class(self, matched_class: MatchedClass) -> str:
        result = StringIO()
        new_class = matched_class.new.to_java_representation()[1:-1]
        original_class = matched_class.original.to_java_representation()[1:-1]
        _ = result.write(f"CLASS {new_class} {original_class}\n")
        for field in matched_class.matched_fields:
            _ = result.write(self.convert_field(field))
        for method in matched_class.matched_methods:
            _ = result.write(self.convert_method(method))
        return result.getvalue()

    @override
    def convert(self, matched_classes: dict[str, MatchedClass]) -> str:
        final = StringIO()
        for matched_class in matched_classes.values():
            _ = final.write(self.convert_class(matched_class))
        return final.getvalue()


class EnigmaParser(Parser):
    def _parse_class(self, components: list[str]) -> MatchedClass:
        new_class = Class.from_java_representation(f"L{components[1]};")
        original_class = Class.from_java_representation(f"L{components[-1]};")
        return MatchedClass(new=new_class, original=original_class, matched_methods=[], matched_fields=[])

    def _parse_field(self, components: list[str]) -> MatchedField:
        new_field = Field(name=components[1], type=components[-1])
        original_field = Field(name=components[-2], type="")
        return MatchedField(new=new_field, original=original_field)

    def _parse_method(self, components: list[str]) -> MatchedMethod:
        method_types = components[-1]
        argument_type, return_type = method_types[1:].split(")")

        new_method = Method(name=components[1], argument_types=argument_type, return_type=return_type)
        original_method = Method(name=components[-2], argument_types="", return_type="")
        return MatchedMethod(new=new_method, original=original_method)

    @override
    def parse(self, raw_input: str) -> dict[str, MatchedClass]:
        result: dict[str, MatchedClass] = {}
        matched_class: MatchedClass | None = None
        current_class = ""
        for line in raw_input.splitlines():
            components = line.split()
            if components[0] == "CLASS":
                matched_class = self._parse_class(components)
                current_class = matched_class.original.name
                result[current_class] = matched_class

            elif components[0] == "FIELD":
                result[current_class].matched_fields.append(self._parse_field(components))

            elif components[0] == "METHOD":
                result[current_class].matched_methods.append(self._parse_method(components))

        return result


class JadxNodeRef(pydantic.BaseModel):
    ref_type: Literal["CLASS", "FIELD", "METHOD"]
    decl_class: str
    short_id: str | None = None

    model_config: ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(
        alias_generator=pydantic.alias_generators.to_camel, validate_by_name=True
    )


class JadxRename(pydantic.BaseModel):
    new_name: str
    node_ref: JadxNodeRef

    model_config: ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(
        alias_generator=pydantic.alias_generators.to_camel, validate_by_name=True
    )


class JadxCodeData(pydantic.BaseModel):
    renames: list[JadxRename]

    model_config: ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(
        alias_generator=pydantic.alias_generators.to_camel, validate_by_name=True
    )


class JadxProjectFile(pydantic.BaseModel):
    code_data: JadxCodeData

    model_config: ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(
        alias_generator=pydantic.alias_generators.to_camel, validate_by_name=True
    )


class JadxFormatter(Formatter):
    @override
    def convert(self, matched_classes: dict[str, MatchedClass]) -> str:
        jadx_project = JadxProjectFile(code_data=JadxCodeData(renames=[]))
        for matched_class in matched_classes.values():
            class_node_ref = JadxNodeRef(ref_type="CLASS", decl_class=matched_class.new.to_full_name())
            class_rename = JadxRename(new_name=matched_class.original.to_full_name(), node_ref=class_node_ref)
            jadx_project.code_data.renames.append(class_rename)
            for matched_field in matched_class.matched_fields:
                field_node_ref = JadxNodeRef(
                    ref_type="FIELD",
                    decl_class=matched_class.new.to_full_name(),
                    short_id=matched_field.new.to_java_representation(),
                )
                field_rename = JadxRename(new_name=matched_field.original.name, node_ref=field_node_ref)
                jadx_project.code_data.renames.append(field_rename)
            for method in matched_class.matched_methods:
                method_node_ref = JadxNodeRef(
                    ref_type="METHOD",
                    decl_class=matched_class.new.to_full_name(),
                    short_id=method.new.to_java_representation(),
                )
                method_rename = JadxRename(new_name=method.original.name, node_ref=method_node_ref)
                jadx_project.code_data.renames.append(method_rename)
        return jadx_project.model_dump_json(indent=4, exclude_unset=True)


class JadxParser(Parser):
    def _parse_class(self, jadx_rename: JadxRename) -> MatchedClass:
        return MatchedClass(
            new=Class.from_full_name(jadx_rename.node_ref.decl_class),
            original=Class.from_full_name(jadx_rename.new_name),
            matched_methods=[],
            matched_fields=[],
        )

    def _parse_holder_class(self, decl_class: str) -> MatchedClass:
        clazz = Class.from_full_name(decl_class)
        return MatchedClass(new=clazz, original=clazz, matched_methods=[], matched_fields=[])

    def _parse_fields(
        self,
        result: dict[str, MatchedClass],
        jadx_to_sigma_classes: dict[str, MatchedClass],
        jadx_to_sigma_field: list[tuple[str, MatchedField]],
    ) -> None:
        for decl_class, matched_field in jadx_to_sigma_field:
            matched_class = jadx_to_sigma_classes.get(decl_class)
            if matched_class is None:
                matched_class = self._parse_holder_class(decl_class)
                result[matched_class.original.name] = matched_class
                jadx_to_sigma_classes[decl_class] = matched_class
            matched_class.matched_fields.append(matched_field)

    def _parse_field(self, jadx_rename: JadxRename) -> tuple[str, MatchedField]:
        assert jadx_rename.node_ref.short_id is not None
        new_field = Field.from_java_representation(jadx_rename.node_ref.short_id)
        original_field = Field(name=jadx_rename.new_name, type="")
        return (jadx_rename.node_ref.decl_class, MatchedField(new=new_field, original=original_field))

    def _parse_methods(
        self,
        result: dict[str, MatchedClass],
        jadx_to_sigma_classes: dict[str, MatchedClass],
        jadx_to_sigma_method: list[tuple[str, MatchedMethod]],
    ) -> None:
        for decl_class, matched_method in jadx_to_sigma_method:
            matched_class = jadx_to_sigma_classes.get(decl_class)
            if matched_class is None:
                matched_class = self._parse_holder_class(decl_class)
                result[matched_class.original.name] = matched_class
                jadx_to_sigma_classes[decl_class] = matched_class
            matched_class.matched_methods.append(matched_method)

    def _parse_method(self, jadx_rename: JadxRename) -> tuple[str, MatchedMethod]:
        assert jadx_rename.node_ref.short_id is not None
        new_method = Method.from_java_representation(jadx_rename.node_ref.short_id)
        original_method = Method(name=jadx_rename.new_name, argument_types="", return_type="")
        return (jadx_rename.node_ref.decl_class, MatchedMethod(new=new_method, original=original_method))

    @override
    def parse(self, raw_input: str) -> dict[str, MatchedClass]:
        result: dict[str, MatchedClass] = {}
        jadx_project = JadxProjectFile.model_validate_json(raw_input)

        jadx_to_sigma_classes: dict[str, MatchedClass] = {}
        jadx_to_sigma_field: list[tuple[str, MatchedField]] = []
        jadx_to_sigma_method: list[tuple[str, MatchedMethod]] = []

        for rename in jadx_project.code_data.renames:
            if rename.node_ref.ref_type == "CLASS":
                matched_class = self._parse_class(rename)
                result[matched_class.original.name] = matched_class
                jadx_to_sigma_classes[rename.node_ref.decl_class] = matched_class
            elif rename.node_ref.ref_type == "FIELD":
                jadx_to_sigma_field.append(self._parse_field(rename))
            elif rename.node_ref.ref_type == "METHOD":
                jadx_to_sigma_method.append(self._parse_method(rename))

        self._parse_fields(result, jadx_to_sigma_classes, jadx_to_sigma_field)
        self._parse_methods(result, jadx_to_sigma_classes, jadx_to_sigma_method)
        return result


class MappingFormat(str, enum.Enum):
    RAW = "raw"
    ENIGMA = "enigma"
    JADX = "jadx"
    LEGACY = "legacy"


FORMAT_TO_FORMATTER: dict[MappingFormat, Formatter] = {
    MappingFormat.RAW: RawFormatter(),
    MappingFormat.ENIGMA: EnigmaFormatter(),
    MappingFormat.JADX: JadxFormatter(),
    MappingFormat.LEGACY: LegacyFormatter(),
}


def convert_to_format(matched_classes: dict[str, MatchedClass], output_format: MappingFormat) -> str:
    try:
        return FORMAT_TO_FORMATTER[output_format].convert(matched_classes)
    except KeyError:
        raise ValueError(
            f"The provided output format is not supported yet: {output_format}."
            + f"Supported formats are: {', '.join(FORMAT_TO_FORMATTER)}"
        ) from None


FORMAT_TO_PARSER: dict[MappingFormat, Parser] = {
    MappingFormat.RAW: RawParser(),
    MappingFormat.ENIGMA: EnigmaParser(),
    MappingFormat.JADX: JadxParser(),
}


def parse_from_format(raw_input: str, input_format: MappingFormat) -> dict[str, MatchedClass]:
    try:
        return FORMAT_TO_PARSER[input_format].parse(raw_input)
    except KeyError:
        raise ValueError(
            f"The provided input format is not supported yet: {input_format}."
            + f"Supported formats are: {', '.join(FORMAT_TO_PARSER)}"
        ) from None



import re
import subprocess
from collections.abc import Iterable
from pathlib import Path


def parse_rg_line(line: str) -> tuple[Path, int]:
    path, _, count = line.rpartition(":")
    return Path(path), int(count)


def rip_regex(pattern: str | re.Pattern[str], search_paths: Iterable[Path]) -> dict[Path, int]:
    if isinstance(pattern, re.Pattern):
        pattern = pattern.pattern
    str_search_paths = [str(path) for path in search_paths]
    args = [
        "rg",
        "--count-matches",
        "--multiline",
        "--no-ignore",
        "--hidden",
        "--regexp",
        pattern,
        *str_search_paths,
    ]
    process = subprocess.run(args, stdout=subprocess.PIPE, text=True, check=False)

    if bool(process.returncode):
        return {}

    if len(str_search_paths) == 1 and (path := next(iter(search_paths))).is_file():
        # if the search path is a single file, ripgrep will not print the filename
        return {path: int(process.stdout)}

    return dict(parse_rg_line(line) for line in process.stdout.splitlines())



import dataclasses
import graphlib
import sys
from abc import ABC, abstractmethod
from collections.abc import Callable, Iterable, Sequence
from functools import cache
from pathlib import Path

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

from sigmatcher.definitions import (
    ClassDefinition,
    Definition,
    ExportDefinition,
    FieldDefinition,
    MacroStatement,
    MethodDefinition,
    Signature,
    SignatureMatch,
)
from sigmatcher.exceptions import (
    DependencyMatchError,
    InvalidMacroModifierError,
    NoMatchesError,
    NoSignaturesError,
    SigmatcherError,
    TooManyMatchesError,
    TooManySignaturesError,
)
from sigmatcher.results import (
    Class,
    Field,
    MatchedClass,
    MatchedExport,
    MatchedField,
    MatchedMethod,
    Method,
    Result,
)


def filter_signature_matches(
    signatures: Iterable[Signature],
    initial_matches: Iterable[SignatureMatch],
    check_signature_callback: Callable[[Signature, set[SignatureMatch]], list[SignatureMatch]],
) -> set[SignatureMatch]:
    all_matches: set[SignatureMatch] = set(initial_matches)
    for signature in signatures:
        signature_match = check_signature_callback(signature, all_matches)
        all_matches.intersection_update(signature_match)
        if len(all_matches) == 0:
            break

    return all_matches


@cache
def get_smali_files(search_root: Path) -> frozenset[Path]:
    return frozenset(search_root.rglob("*.smali"))


@dataclasses.dataclass(frozen=True)
class Analyzer(ABC):
    definition: Definition
    app_version: str | None

    @abstractmethod
    def analyze(self, results: dict[str, Result | SigmatcherError]) -> Result:
        pass

    def check_match_count(
        self, matches: set[SignatureMatch] | None, signatures: tuple[Signature, ...] | None = None
    ) -> None:
        if matches is None or len(matches) == 0:
            raise NoMatchesError(self.name, signatures)
        if len(matches) > 1:
            raise TooManyMatchesError[SignatureMatch](self.name, matches, signatures)

    def get_dependencies(self) -> set[str]:
        return self.definition.get_dependencies(self.app_version)

    def check_dependencies(self, results: dict[str, Result | SigmatcherError]) -> None:
        failed_dependencies: list[str] = []
        for dependency_name in self.get_dependencies():
            child_result = results[dependency_name]
            if isinstance(child_result, Exception):
                failed_dependencies.append(dependency_name)

        if failed_dependencies:
            raise DependencyMatchError(self.name, failed_dependencies)

    def resolve_macro(self, results: dict[str, Result | SigmatcherError], macro_statement: MacroStatement) -> str:
        result = results[macro_statement.subject]
        assert not isinstance(result, Exception)

        try:
            resolved_macro = getattr(result.new, macro_statement.modifier)  # pyright: ignore[reportAny]
        except AttributeError:
            raise InvalidMacroModifierError(self.name, macro_statement, result.new.__class__.__name__) from None

        assert isinstance(resolved_macro, str)
        return resolved_macro

    def get_resolved_signatures(self, results: dict[str, Result | SigmatcherError]) -> tuple[Signature, ...]:
        resolved_signatures: list[Signature] = []
        for signature in self.get_signatures_for_version():
            resolved_signature = signature
            for macro_statement in signature.get_macro_definitions():
                resolved_macro = self.resolve_macro(results, macro_statement)
                resolved_signature = resolved_signature.resolve_macro(macro_statement, resolved_macro)
            resolved_signatures.append(resolved_signature)

        return tuple(resolved_signatures)

    def get_signatures_for_version(self) -> tuple[Signature, ...]:
        return self.definition.get_signatures_for_version(self.app_version)

    @property
    def name(self) -> str:
        return self.definition.name

    @override
    def __repr__(self) -> str:
        return self.name


@dataclasses.dataclass(frozen=True)
class ClassAnalyzer(Analyzer):
    definition: ClassDefinition
    search_root: Path

    @override
    def analyze(self, results: dict[str, Result | SigmatcherError]) -> MatchedClass:
        signatures = list(self.get_resolved_signatures(results))
        if len(signatures) == 0:
            raise NoSignaturesError(self.name)

        # Make sure the first signature is a whitelist signature in order to improve performance
        whitelist_signature_index = 0
        for i, signature in enumerate(signatures):
            if signature.count.min_count > 0:
                whitelist_signature_index = i
                break
        signatures.insert(0, signatures.pop(whitelist_signature_index))

        def check_signature_callback(signature: Signature, matches: set[Path]) -> list[Path]:
            return signature.check_files(matches, self.search_root)

        initial_matches = get_smali_files(self.search_root)
        class_matches = filter_signature_matches(signatures, initial_matches, check_signature_callback)

        self.check_match_count(class_matches, tuple(signatures))
        match = next(iter(class_matches))
        with match.open() as f:
            class_definition_line = f.readline().rstrip("\n")
        _, _, raw_class_name = class_definition_line.rpartition(" ")
        new_class = Class.from_java_representation(raw_class_name)
        original_class = Class(name=self.definition.name, package=self.definition.package or new_class.package)
        return MatchedClass(
            original=original_class, new=new_class, smali_file=match, matched_methods=[], matched_fields=[]
        )


@dataclasses.dataclass(frozen=True)
class FieldAnalyzer(Analyzer):
    definition: FieldDefinition
    parent: ClassAnalyzer

    @override
    def analyze(self, results: dict[str, Result | SigmatcherError]) -> MatchedField:
        parent_class_result = results[self.parent.name]
        assert isinstance(parent_class_result, MatchedClass)
        assert isinstance(parent_class_result.smali_file, Path)

        signatures = self.get_resolved_signatures(results)
        if len(signatures) == 0:
            raise NoSignaturesError(self.name)
        if len(signatures) > 1:
            raise TooManySignaturesError(self.name, signatures)
        signature = signatures[0]

        raw_class = parent_class_result.smali_file.read_text()
        captured_names = signature.capture(raw_class)
        self.check_match_count(captured_names, signatures)
        raw_field_name = next(iter(captured_names))

        new_field = Field.from_java_representation(raw_field_name)
        # TODO: should we get the types for the original field from the definition?
        original_field = Field(name=self.definition.name, type=new_field.type)
        matched_field = MatchedField(original=original_field, new=new_field)
        parent_class_result.matched_fields.append(matched_field)
        return matched_field

    @override
    def get_dependencies(self) -> set[str]:
        return super().get_dependencies() | {self.parent.name}

    @property
    @override
    def name(self) -> str:
        return f"{self.parent.name}.fields.{self.definition.name}"


@dataclasses.dataclass(frozen=True)
class MethodAnalyzer(Analyzer):
    definition: MethodDefinition
    parent: ClassAnalyzer

    @override
    def analyze(self, results: dict[str, Result | SigmatcherError]) -> MatchedMethod:
        parent_class_result = results[self.parent.name]
        assert isinstance(parent_class_result, MatchedClass)
        assert isinstance(parent_class_result.smali_file, Path)

        raw_methods = parent_class_result.smali_file.read_text().split(".method")[1:]
        methods = {".method" + method for method in raw_methods}

        signatures = self.get_resolved_signatures(results)
        if len(signatures) == 0:
            raise NoSignaturesError(self.name)

        def check_signature_callback(signature: Signature, matches: set[str]) -> list[str]:
            return signature.check_strings(matches)

        method_matches = filter_signature_matches(signatures, methods, check_signature_callback)

        self.check_match_count(method_matches, signatures)
        match = next(iter(method_matches))
        method_definition_line, _, _ = match.partition("\n")
        _, _, raw_method_name = method_definition_line.rpartition(" ")

        new_method = Method.from_java_representation(raw_method_name)
        # TODO: should we get the types for the original method from the definition?
        original_method = Method(
            name=self.definition.name, argument_types=new_method.argument_types, return_type=new_method.return_type
        )
        matched_method = MatchedMethod(original=original_method, new=new_method)
        parent_class_result.matched_methods.append(matched_method)
        return matched_method

    @override
    def get_dependencies(self) -> set[str]:
        return super().get_dependencies() | {self.parent.name}

    @property
    @override
    def name(self) -> str:
        return f"{self.parent.name}.methods.{self.definition.name}"


@dataclasses.dataclass(frozen=True)
class ExportAnalyzer(Analyzer):
    definition: ExportDefinition
    parent: ClassAnalyzer

    @override
    def analyze(self, results: dict[str, Result | SigmatcherError]) -> MatchedExport:
        parent_class_result = results[self.parent.name]
        assert isinstance(parent_class_result, MatchedClass)
        assert isinstance(parent_class_result.smali_file, Path)

        signatures = self.get_resolved_signatures(results)
        if len(signatures) == 0:
            raise NoSignaturesError(self.name)
        if len(signatures) > 1:
            raise TooManySignaturesError(self.name, signatures)
        signature = signatures[0]

        raw_class = parent_class_result.smali_file.read_text()
        captured_names = signature.capture(raw_class)
        self.check_match_count(captured_names, signatures)
        export_value = next(iter(captured_names))

        return MatchedExport.from_value(export_value)

    @override
    def get_dependencies(self) -> set[str]:
        return super().get_dependencies() | {self.parent.name}

    @property
    @override
    def name(self) -> str:
        return f"{self.parent.name}.exports.{self.definition.name}"


def create_analyzers(
    definitions: Sequence[ClassDefinition], search_root: Path, app_version: str | None
) -> dict[str, Analyzer]:
    canonical_search_root = search_root.resolve()
    name_to_analyzer: dict[str, Analyzer] = {}
    for class_definition in definitions:
        if not class_definition.is_in_version_range(app_version):
            continue

        class_analyzer = ClassAnalyzer(class_definition, app_version, canonical_search_root)
        name_to_analyzer[class_definition.name] = class_analyzer

        for method_definition in class_definition.methods:
            if not method_definition.is_in_version_range(app_version):
                continue
            method_analyzer = MethodAnalyzer(method_definition, app_version, parent=class_analyzer)
            name_to_analyzer[method_analyzer.name] = method_analyzer

        for field_definition in class_definition.fields:
            if not field_definition.is_in_version_range(app_version):
                continue
            field_analyzer = FieldAnalyzer(field_definition, app_version, parent=class_analyzer)
            name_to_analyzer[field_analyzer.name] = field_analyzer

        for export_definition in class_definition.exports:
            if not export_definition.is_in_version_range(app_version):
                continue
            export_analyzer = ExportAnalyzer(export_definition, app_version, parent=class_analyzer)
            name_to_analyzer[export_analyzer.name] = export_analyzer

    return name_to_analyzer


def analyze(
    definitions: Sequence[ClassDefinition], unpacked_path: Path, app_version: str | None
) -> dict[str, Result | SigmatcherError]:
    name_to_analyzer = create_analyzers(definitions, unpacked_path, app_version)

    sorter: graphlib.TopologicalSorter[str] = graphlib.TopologicalSorter()
    for analyzer in name_to_analyzer.values():
        sorter.add(analyzer.name, *analyzer.get_dependencies())

    results: dict[str, Result | SigmatcherError] = {}
    for analyzer_name in sorter.static_order():
        analyzer = name_to_analyzer[analyzer_name]
        try:
            analyzer.check_dependencies(results)
            results[analyzer_name] = analyzer.analyze(results)
        except SigmatcherError as e:
            results[analyzer_name] = e
    return results



from pathlib import Path
from typing import TypeAlias

import pydantic


class Export(pydantic.BaseModel):
    value: str


class MatchedExport(pydantic.BaseModel):
    new: Export

    @classmethod
    def from_value(cls, value: str) -> "MatchedExport":
        return cls(new=Export(value=value))


class Field(pydantic.BaseModel):
    name: str
    type: str

    @classmethod
    def from_java_representation(cls, java_representation: str) -> "Field":
        name, _, field_type = java_representation.partition(":")
        return cls(name=name, type=field_type)

    def to_java_representation(self) -> str:
        return f"{self.name}:{self.type}"

    @property
    def java(self) -> str:
        return self.to_java_representation()


class MatchedField(pydantic.BaseModel):
    original: Field
    new: Field


class Method(pydantic.BaseModel):
    name: str
    argument_types: str
    return_type: str

    @classmethod
    def from_java_representation(cls, java_representation: str) -> "Method":
        name, _, types = java_representation.partition("(")
        argument_types, _, return_type = types.partition(")")
        return cls(name=name, argument_types=argument_types, return_type=return_type)

    def to_java_representation(self) -> str:
        return f"{self.name}({self.argument_types}){self.return_type}"

    @property
    def java(self) -> str:
        return self.to_java_representation()


class MatchedMethod(pydantic.BaseModel):
    original: Method
    new: Method


class Class(pydantic.BaseModel):
    name: str
    package: str

    @classmethod
    def from_full_name(cls, full_name: str) -> "Class":
        package, _, name = full_name.rpartition(".")
        return cls(name=name, package=package)

    def to_full_name(self) -> str:
        return f"{self.package}.{self.name}"

    @property
    def full_name(self) -> str:
        return self.to_full_name()

    @classmethod
    def from_java_representation(cls, java_representation: str) -> "Class":
        package, _, name = java_representation[1:-1].replace("/", ".").rpartition(".")
        return cls(name=name, package=package)

    def to_java_representation(self) -> str:
        return f"L{self.to_full_name()};".replace(".", "/")

    @property
    def java(self) -> str:
        return self.to_java_representation()


class MatchedClass(pydantic.BaseModel):
    original: Class
    new: Class
    matched_methods: list[MatchedMethod]
    matched_fields: list[MatchedField]
    smali_file: Path | None = pydantic.Field(default=None, exclude=True)


Result: TypeAlias = MatchedClass | MatchedField | MatchedMethod | MatchedExport
